/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#include "inc_vendor.cl"
#include "inc_types.cl"
#include "inc_common.cl"

__kernel void m02000_mxx (KERN_ATTR_BASIC ())
{
}

__kernel void m02000_sxx (KERN_ATTR_BASIC ())
{
}
