/**
 * Author......: See docs/credits.txt
 * License.....: MIT
 */

#ifndef _USAGE_H
#define _USAGE_H

#include <stdio.h>
#include <string.h>

void usage_mini_print (const char *progname);
void usage_big_print  (const char *progname);

#endif // _USAGE_H
